--[=====[



--]=====]

local class = require('Middleclass').class

-- local default_image = love.graphics.newImage('error.png')


local Entity = class('Entity')

function Entity:initialize(type, x, y, angle)
    print(string.format('new Entity: %s %s %s %s', type,x,y,angle))
    self:position(x,y,angle)
    self.type = type
    self:setup(type) --note runs after default values
    self.image = default_image
end
Entity.x = 0
Entity.y = 0
Entity.angle = 0
Entity.radius = 32

Entity.shape = 'circle'
Entity.width = 64
Entity.height = 64


function Entity:setup(string) --string is the 'type'
    if not string or string == '' then
        print('WARNING, empty setup string')

    elseif string == 'test' then
        self.angle = math.rad(45)

    elseif string == 'test2' then
        self.angle = math.rad(45)
        self.shape = 'rectangle'

    elseif string == 'platform' then
        self.physics_body_type = 'static'
        self.shape = 'rectangle'
        self.width = 64*8
        self.height = 64

    elseif string == 'static' then
        self.physics_body_type = 'static'
        self.shape = 'rectangle'

    else 
        print('WARNING, unrecognised setup string: ' .. string)
    end
end


Entity.mass = 1
Entity.physics_linear_damping = 0
Entity.restitution = 0.4

Entity.physics_body_type = 'dynamic'

Entity.physics_fixture_density = 1

Entity.fixed_rotation = false

function Entity:initialize_physics()

    --STUFF WE MIGHT WANT TO CHANGE REALTIME:
    --position angle done
    --size?!?! bouncy, strip project to remove bug

    if not self.physics_object then
        self.physics_object = {} --create a container for the physics_data
        self.physics_object.b = love.physics.newBody(world, self.x,self.y, self.physics_body_type)
        self.physics_object.b:setMass(self.mass)

        if self.shape == 'circle' then --circle
            print('circle')
            self.physics_object.s = love.physics.newCircleShape(self.radius)
            self.circle = true --booleans are quicker to check on draw
        end
        if self.shape == 'rectangle' then --rectangle
            self.physics_object.s = love.physics.newRectangleShape(self.width,self.height)
            self.rectangle = true
        end


        self.physics_object.b:setLinearDamping(self.physics_linear_damping,self.physics_linear_damping) --like drag

        self.physics_object.f = love.physics.newFixture(self.physics_object.b, self.physics_object.s, self.physics_fixture_density)
        self:setRestitution(self.restitution)
        self.physics_object.f:setUserData(self)

        self.physics_object.b:setFixedRotation(self.fixed_rotation) --fix rotation

        self:setAngle(self.angle)


        self:setSensor(self.physics_sensor )


        self:setFilterData(
            self.physics_category_list,
            self.physics_mask_list,
            self.physics_group
            )
    end
end

Entity.physics_group = 1



function Entity:setFilterData(categories, mask, group)
        self:setFixtureCategory(categories)
        self:setFixtureMask(mask)
        self:setFixtureGroupIndex(group)
end


Entity.physics_category_list = 1
function Entity:setFixtureCategory(...)
    if self.physics_object then
        self.physics_object.f:setCategory(...)
    else
        self.physics_category_list = {...} --save as table, need to load after physics
    end
end


Entity.physics_mask_list = 1
function Entity:setFixtureMask(...)
    if self.physics_object then
        self.physics_object.f:setMask(...)
    else
        self.physics_mask_list = {...} --save as table, need to load after physics
    end
end

function Entity:setFixtureGroupIndex(group)
    if self.physics_object then
        self.physics_object.f:setGroupIndex(group)
    else
       self.physics_group = group
    end
end

function Entity:setSensor(bool)
    if self.physics_object then
        self.physics_object.f:setSensor(bool)
    else
        self.physics_sensor = bool
    end
end

function Entity:getSensor()
    if self.physics_object then
        return self.physics_object.f:isSensor()
    end
end


function Entity:setRestitution(restitution)
    -- self.physics_object.f:setRestitution(0.4)
    -- restitution = restitution or self.physics_restitution
    if self.physics_object then
        self.physics_object.f:setRestitution(restitution)
    else
        self.restitution = restitution
    end
end



function Entity:getX()
    if self.physics_object then
        return self.physics_object.b:getX()
    else
        return self.x --if no physics_object
    end
end
function Entity:getY()
    if self.physics_object then
        return self.physics_object.b:getY()
    else
        return self.y --if no physics_object
    end
end
function Entity:setX(x)
    if self.physics_object then
        self.physics_object.b:setX(x)
    else
        self.x = x --if no physics_object
    end
end
function Entity:setY(y)
    if self.physics_object then
        self.physics_object.b:setY(y)
    else
        self.y = y --if no physics_object
    end
end
function  Entity:getAngle(angle)
    if self.physics_object then
        return self.physics_object.b:getAngle()
    else
        return self.angle --if no physics_object
    end
end
function  Entity:setAngle(angle)
    if self.physics_object then
        self.physics_object.b:setAngle(angle)
    else
        self.angle = angle --if no physics_object
    end
end
function Entity:position(x, y , angle)
    --EXERIMENTAL
    --can get and set all position pars
    --maybe slow if you use this to set pars
    if x then
        self:setX(x)
    end
    if y then
        self:setY(y)
    end
    if angle then
        self:setAngle(angle)
    end
    return self:getX(), self:getY(), self:getAngle()
end


function Entity:update(dt)
end




function Entity:draw_physics_shape() --debug function visualises physics shapes, too slow for gameplay 
    
    if self.physics_object then --if we have physics

        if self.physics_object.f:isSensor() then --if it is a sensor, highlight it orange
            love.graphics.setColor(255,127,0)
        else
            love.graphics.setColor(255,0,0) --if not, highlight red
        end

        local physics_shape = self.physics_object.s:type() --check for shape (slow) and draw based on shape:
        if physics_shape == 'CircleShape' then
            love.graphics.circle("line", self.physics_object.b:getX(),self.physics_object.b:getY(), self.physics_object.s:getRadius())
        elseif physics_shape == 'PolygonShape' then
            love.graphics.polygon("line", self.physics_object.b:getWorldPoints(self.physics_object.s:getPoints()))
        end
    end

end

Entity.red = 0
Entity.green = 255
Entity.blue = 255
Entity.alpha = 255


function Entity:draw()

    love.graphics.setColor(
        self.red,
        self.green,
        self.blue,
        self.alpha)


    local x,y,angle = self:getX(),self:getY(),self:getAngle()
    -- love.graphics.rectangle('line', x, y, self.width, self.height)
    --debug circle with line facing direction
    love.graphics.circle('line', x,y,self.radius)
    local x_dir, y_dir = math.angle_to_vector(self:getAngle())
    love.graphics.line(x, y, x + x_dir * self.radius, y+y_dir * self.radius)

    if debug_draw_physics_shapes then
        self:draw_physics_shape() -- DEBUG
    end

end

function Entity:lightdraw()

    local x,y,angle = self:getX(),self:getY(),self:getAngle()

    love.graphics.setColor(255,255,255,255)

    -- love.graphics.rectangle('line', x-(self.width/2), y-(self.height/2), self.width, self.height)
    --debug circle with line facing direction
    love.graphics.circle('fill', x,y,self.radius)
    -- local x_dir, y_dir = math.angle_to_vector(self:getAngle())
    -- love.graphics.line(x, y, x + x_dir * self.radius, y+y_dir * self.radius)

    if self.rectangle then
        love.graphics.rectangle('fill', x-(self.width/2),y-(self.height/2),self.width,self.height)
    end


    local image_width, image_height = self.image:getWidth(), self.image:getHeight()
    local x_scale, y_scale = self.radius / image_width * 2, self.radius / image_height * 2

    love.graphics.draw(
        self.image,
        x, --position
        y,
        angle, --rotation
        x_scale, --scale
        y_scale,
        image_width/2,
        image_height/2)

end


return Entity