local class = require('Middleclass').class
local csv = require('csv')
local json = require('json')

local LightWorld = require('lightworld')

default_image = love.graphics.newImage('error.png') --deliberate global!



-- local Action = require('ee_action')

local Engine = require('ee_engine')
-- local Entity = require('ee_entity')


debug_draw_physics_shapes = true --causes physics to be show



grid = {}
grid.scale = 64


function math.normalize(x,y) local l=(x*x+y*y)^.5 if l==0 then return 0,0,0 else return x/l,y/l,l end end
function math.vector_to_angle(x,y) return math.atan2(x, -y) end
function math.angle_to_vector(angle) return math.sin(angle), -math.cos(angle) end
function math.round(number) return math.floor(number+0.5) end

function grid_coor_to_pos(x,y)
    x,y = math.round(x) * grid.scale, math.round(y) * grid.scale
    return x,y
end

function pos_to_grid_coor(x,y)
    x, y = math.round(x/grid.scale),math.round(y/grid.scale)
    return x,y
end

function snap_to_grid(x,y)
    return grid_coor_to_pos(pos_to_grid_coor(x,y))
end



Camera = class('Camera')
Camera.x, Camera.y = 0, 0
-- Camera.x_scale, Camera.y_scale = 1,1
Camera.scale = 1
Camera.rotation = 0
Camera.speed = 500
Camera.zoom_speed = 1
Camera.n_key = 'w'
Camera.e_key = 'd'
Camera.s_key = 's'
Camera.w_key = 'a' --zoom_in_key
Camera.zoom_in_key = 'e'
Camera.zoom_out_key = 'q'
Camera.rotate_right_key = 'r'
Camera.rotate_left_key = 'f'

function Camera:zoom(dt, val)
    -- self.x_scale = val * self.zoom_speed * dt + self.x_scale
    -- self.y_scale = val * self.zoom_speed * dt + self.y_scale
    self.scale = val * self.zoom_speed * dt + self.scale
end
function Camera:zoom_in(dt)
    self:zoom(dt,1)
end
function Camera:zoom_out(dt)
    self:zoom(dt,-1)
end
function Camera:rotate(dt, val)
    self.rotation = val * self.zoom_speed * dt + self.rotation
end
function Camera:rotate_right(dt)
    self:rotate(dt,1)
end
function Camera:rotate_left(dt)
    self:rotate(dt,-1)
end



camera = Camera --GLOBAL, used by engine



local mouse = {}
mouse.x,mouse.y = 0,0

local screen = {}
screen.width,screen.height = love.graphics.getWidth(), love.graphics.getHeight()


local gravity = {}
gravity.x, gravity.y = 0, 0
gravity.x, gravity.y = 0, 9.8 * grid.scale /8


local grid = {}
grid.scale = 64





--GameObject and GameManager inherit these functions
SharedClass = class('SharedClass')
--Event Lambdas
function SharedClass:add_update_event(funct) --update evenst run each update, used by GameObject and GameManager
    if not self.update_events then
        self.update_events = Event()
    end
    self.update_events:add_function(funct)
end
function SharedClass:run_update_events(dt)
    if self.update_events then
        self.update_events:invoke(dt)
    end
end
function SharedClass:add_collect_event(funct) --update evenst run each update, used by GameObject and GameManager
    if not self.collect_events then
        self.collect_events = Event()
    end
    self.collect_events:add_function(funct)
end
function SharedClass:run_collect_events(collected_object)
    if self.collect_events then
        self.collect_events:invoke(collected_object)
    end
end
function SharedClass:add_collected_event(funct) --update evenst run each update, used by GameObject and GameManager
    if not self.collected_events then
        self.collected_events = Event()
    end
    self.collected_events:add_function(funct)
end
function SharedClass:run_collected_events(collected_object)
    if self.collected_events then
        self.collected_events:invoke(collected_object)
    end
end
function SharedClass:add_beginContact_event(funct) --begin contact events are for physics, only used by GameObject
    if not self.beginContact_events then
        self.beginContact_events = Event()
    end
    self.beginContact_events:add_function(funct)
end
function SharedClass:run_beginContact_events(collision_object, coll)
    --collision_object is the other object we hit, like we are ball we hit wall
    if self.beginContact_events then
        self.beginContact_events:invoke(collision_object, coll)
    end
end
--EVENT TIMER, add events to fiture
function SharedClass:add_scheduled_event(time_till_event, event_function)
    --event system, allows adding of future events as lambda function
    if not self.scheduled_events then --if no self.events
        self.scheduled_events = {} --add a table
    end
    --insert the event at the end
    --self.events looks like a list of tuples {{event_time, event_function}...}
    table.insert(self.scheduled_events, {time_till_event + love.timer.getTime(), event_function}) 
    --compare function checks the event_times to sort the list of events
    local function compare(a,b) 
        return a[1] < b[1] --if this event is sooner than the compared one
    end
    --sorts events from soonest (first), to latest (last)
    table.sort(self.scheduled_events, compare) 
end
function  SharedClass:print_scheduled_events()
    --for debug, print pending events
    if self.scheduled_events then
        for n,v in ipairs(self.scheduled_events) do --debug print
            print (v[1] ..', ' .. tostring(v[2]))
        end
    end
end
function SharedClass:update_scheduled_events()
    --run on update, to run all due events
    if self.scheduled_events then
        if #self.scheduled_events > 0 then
            local current_event = self.scheduled_events[1] --first is start of list
            local event_time = current_event[1]
            local event_lambda = current_event[2]
            if love.timer.getTime() >= current_event[1] then
                if type(event_lambda) == 'string' then
                    print('event string: ' .. event_lambda)
                elseif type(event_lambda) == 'function' then
                    event_lambda(dt)
                end
                table.remove(self.scheduled_events,1) --remove the triggered event from the start of the list
                self:update_scheduled_events() --try a recurse, incase there are other events at the same time
            end
        end
    end
end












local ambient_light = 127
local ambient_light = 50




function love.load()
    lightWorld = LightWorld({ --GLOBAL, deliberate!
        ambient = {ambient_light,ambient_light,ambient_light},
        })

    lightWorld:newLight(100, 100, 127, 127, 127, 200) -- test light


    engine = Engine()

    world = love.physics.newWorld(gravity.x, gravity.y, true)
    world:setCallbacks(beginContact, endContact, preSolve, postSolve)

    spawn1 = engine:spawn('test', 20,0)
    spawn2 = engine:spawn('test2', grid_coor_to_pos(1,1))

    -- spawn3 = engine:spawn('static', grid_coor_to_pos(0,4))
    spawn3 = engine:spawn('platform', grid_coor_to_pos(0,4))

    -- spawn1:setSensor(true)

    spawn1:initialize_physics()
    spawn2:initialize_physics()
    spawn3:initialize_physics()

    spawn3 = engine:spawn('test2', grid_coor_to_pos(1,1))



    for n,v in ipairs(engine.entities) do
        print(n,v, v.type)
    end


end

function love.update(dt)

    world:update(dt) --update Love2D physics
    lightWorld:update(dt)
    lightWorld:setTranslation(camera.x, camera.y, camera.scale)

    -- lightWorld:rotate(camera.rotation)

    -- lightWorld:setTranslation(camera.x, camera.y)

    engine:update(dt)

    if love.keyboard.isDown(camera.n_key) then --N
            camera.y = camera.y + (dt * camera.speed)
    end
    if love.keyboard.isDown(camera.e_key) then --E
            camera.x = camera.x - (dt * camera.speed)
    end
    if love.keyboard.isDown(camera.s_key) then --S
            camera.y = camera.y - (dt * camera.speed)
    end
    if love.keyboard.isDown(camera.w_key) then --W
            camera.x = camera.x + (dt * camera.speed)
    end
    if love.keyboard.isDown(camera.zoom_in_key) then --zoom in
            -- camera.x = camera.x - (dt * camera.speed)
            -- camera.x_scale = camera.x_scale + dt
            -- camera.y_scale = camera.y_scale + dt
            camera:zoom_in(dt)
    end
    if love.keyboard.isDown(camera.zoom_out_key) then --zoom out
            -- camera.x = camera.x - (dt * camera.speed)
            -- camera.x_scale = camera.x_scale - dt
            -- camera.y_scale = camera.y_scale - dt
            camera:zoom_out(dt)
    end
    if love.keyboard.isDown(camera.rotate_right_key) then
        camera:rotate_right(dt)
    end
    if love.keyboard.isDown(camera.rotate_left_key) then
        camera:rotate_left(dt)
    end

    

end

background_col = 30

function love.draw()

-- love.graphics.rectangle('fill',0,0,1000,1000)

    -- love.graphics.setBackgroundColor(background_col,background_col,background_col)
    
    


    -- lightWorld:draw(function()
        engine:draw()
      -- end)

end






--Physics hooks
function beginContact(a, b, coll)
    -- print('beginContact!')

    local x, y = coll:getNormal()
    local aUserData, bUserData = a:getUserData(), b:getUserData()

    -- bUserData:beginContact(aUserData, coll) --dynamic that hits static --these need to be added back
    -- aUserData:beginContact(bUserData, coll)

    
    -- print(tostring(aUserData).." colliding with "..tostring(bUserData).." with a vector normal of: "..x..", "..y)
    -- text = text.."\n"..tostring(aUserData).." colliding with "..tostring(bUserData).." with a vector normal of: "..x..", "..y
end
 
function endContact(a, b, coll)
    -- print('endContact!')

    local aUserData, bUserData = a:getUserData(), b:getUserData()
    persisting = 0 --warning global
    -- text = text.."\n"..tostring(aUserData).." uncolliding with "..tostring(bUserData)
end


persisting = 0 --global persisting (BAD)

function preSolve(a, b, coll)
    -- print('preSolve!')

    --MYADDED NOT SURE YET!!!

    local aUserData, bUserData = a:getUserData(), b:getUserData()

    -- bUserData:preSolve(aUserData, coll) --dynamic that hits static
    -- aUserData:preSolve(bUserData, coll)

    --MYADDED END

    --NEED TO TRACK INDIVIDUAL PERSISTING!!!! we currenly have global
    
    if persisting == 0 then    -- only say when they first start touching
        -- text = text.."\n"..tostring(a:getUserData()).." touching "..tostring(b:getUserData())
    elseif persisting < 20 then    -- then just start counting
        -- text = text.." "..persisting
    end
    persisting = persisting + 1    -- keep track of how many updates they've been touching for

    -- print('collPos:',coll:getPositions()) 
    -- print('aPos:',a:getUserData():getX(), a:getUserData():getY())
    -- print('bPos:',b:getUserData():getX(), b:getUserData():getY())

    -- local x1, y1, x2, y2 = coll:getPositions()


    --x1, y1, x2, y2 = Contact:getPositions()
end
 
function postSolve(a, b, coll, normalimpulse, tangentimpulse)
    -- print('postSolve!')

end



